@@ -25,16 +25,15 @@ from .fs.s3 import S3FileSystem
 from .hash_info import HashInfo
 from .istextfile import istextfile
 from .objects import Tree
-from .objects import save as osave
 from .objects.errors import ObjectFormatError
 from .objects.stage import stage as ostage
+from .objects.transfer import transfer as otransfer
 from .scheme import Schemes
 from .utils import relpath
 from .utils.fs import path_isin
 
 if TYPE_CHECKING:
     from .objects.db.base import ObjectDB
-    from .objects.file import HashFile
 
 logger = logging.getLogger(__name__)
 
@@ -404,21 +403,21 @@ class Output:
         return self.odb.hash_to_path_info(self.hash_info.value).url
 
     def get_hash(self):
-        if not self.use_cache:
-            return ostage(
-                self.repo.odb.local,
-                self.path_info,
-                self.fs,
-                self.fs.PARAM_CHECKSUM,
-                dvcignore=self.dvcignore,
-            ).hash_info
-        return ostage(
-            self.odb,
+        if self.use_cache:
+            odb = self.odb
+            name = self.odb.fs.PARAM_CHECKSUM
+        else:
+            odb = self.repo.odb.local
+            name = self.fs.PARAM_CHECKSUM
+        _, obj = ostage(
+            odb,
             self.path_info,
             self.fs,
-            self.odb.fs.PARAM_CHECKSUM,
+            name,
             dvcignore=self.dvcignore,
-        ).hash_info
+            dry_run=not self.use_cache,
+        )
+        return obj.hash_info
 
     @property
     def is_dir_checksum(self):
@@ -545,7 +544,7 @@ class Output:
             logger.debug("Output '%s' didn't change. Skipping saving.", self)
             return
 
-        self.obj = ostage(
+        _, self.obj = ostage(
             self.odb,
             self.path_info,
             self.fs,
@@ -566,14 +565,28 @@ class Output:
         assert self.hash_info
 
         if self.use_cache:
-            obj = ostage(
-                self.odb,
-                filter_info or self.path_info,
-                self.fs,
-                self.odb.fs.PARAM_CHECKSUM,
-                dvcignore=self.dvcignore,
+            granular = (
+                self.is_dir_checksum
+                and filter_info
+                and filter_info != self.path_info
             )
-            objects.save(self.odb, obj)
+            if granular:
+                obj = self._commit_granular_dir(filter_info)
+            else:
+                staging, obj = ostage(
+                    self.odb,
+                    filter_info or self.path_info,
+                    self.fs,
+                    self.odb.fs.PARAM_CHECKSUM,
+                    dvcignore=self.dvcignore,
+                )
+                otransfer(
+                    staging,
+                    self.odb,
+                    {obj.hash_info},
+                    shallow=False,
+                    move=True,
+                )
             checkout(
                 filter_info or self.path_info,
                 self.fs,
@@ -585,6 +598,26 @@ class Output:
             )
             self.set_exec()
 
+    def _commit_granular_dir(self, filter_info):
+        prefix = filter_info.relative_to(self.path_info).parts
+        staging, save_obj = ostage(
+            self.odb,
+            self.path_info,
+            self.fs,
+            self.odb.fs.PARAM_CHECKSUM,
+            dvcignore=self.dvcignore,
+        )
+        save_obj = save_obj.filter(prefix)
+        checkout_obj = save_obj.get(prefix)
+        otransfer(
+            staging,
+            self.odb,
+            {save_obj.hash_info} | {entry.hash_info for _, entry in save_obj},
+            shallow=True,
+            move=True,
+        )
+        return checkout_obj
+
     def dumpd(self):
         ret = copy(self.hash_info.to_dict())
 
@@ -670,7 +703,7 @@ class Output:
 
         if filter_info and filter_info != self.path_info:
             prefix = filter_info.relative_to(self.path_info).parts
-            obj = obj.filter(self.odb, prefix, **kwargs)
+            obj = obj.get(prefix)
 
         return obj
 
@@ -764,7 +797,7 @@ class Output:
 
         upload = not (update and from_fs.isdir(from_info))
         jobs = jobs or min((from_fs.jobs, odb.fs.jobs))
-        obj = ostage(
+        staging, obj = ostage(
             odb,
             from_info,
             from_fs,
@@ -773,7 +806,14 @@ class Output:
             jobs=jobs,
             no_progress_bar=no_progress_bar,
         )
-        osave(odb, obj, jobs=jobs, move=upload)
+        otransfer(
+            staging,
+            odb,
+            {obj.hash_info},
+            jobs=jobs,
+            move=upload,
+            shallow=False,
+        )
 
         self.hash_info = obj.hash_info
         return obj
@@ -802,8 +842,11 @@ class Output:
         obj = self.odb.get(self.hash_info)
         try:
             objects.check(self.odb, obj)
-        except (FileNotFoundError, ObjectFormatError):
-            self.repo.cloud.pull([obj], show_checksums=False, **kwargs)
+        except FileNotFoundError:
+            self.repo.cloud.pull([obj.hash_info], **kwargs)
+
+        if self.obj:
+            return self.obj
 
         try:
             self.obj = objects.load(self.odb, self.hash_info)
@@ -814,8 +857,8 @@ class Output:
 
     def collect_used_dir_cache(
         self, remote=None, force=False, jobs=None, filter_info=None
-    ) -> Dict[Optional["ObjectDB"], Set["HashFile"]]:
-        """Fetch dir cache and return used objects for this out."""
+    ) -> Dict[Optional["ObjectDB"], Set["HashInfo"]]:
+        """Fetch dir cache and return used object IDs for this out."""
 
         try:
             self.get_dir_cache(jobs=jobs, remote=remote)
@@ -824,7 +867,7 @@ class Output:
 
         try:
             objects.check(self.odb, self.odb.get(self.hash_info))
-        except (FileNotFoundError, ObjectFormatError):
+        except FileNotFoundError:
             msg = (
                 "Missing cache for directory '{}'. "
                 "Cache for files inside will be lost. "
@@ -837,14 +880,16 @@ class Output:
                 )
             return {}
 
-        obj = self.get_obj(filter_info=filter_info, copy=True)
-        self._set_obj_names(obj)
-        return {None: {obj}}
+        obj = self.get_obj()
+        if filter_info and filter_info != self.path_info:
+            prefix = filter_info.relative_to(self.path_info).parts
+            obj = obj.filter(prefix)
+        return {None: set(self._named_obj_ids(obj))}
 
     def get_used_objs(
         self, **kwargs
-    ) -> Dict[Optional["ObjectDB"], Set["HashFile"]]:
-        """Return filtered set of used objects for this out."""
+    ) -> Dict[Optional["ObjectDB"], Set["HashInfo"]]:
+        """Return filtered set of used object IDs for this out."""
 
         if not self.use_cache:
             return {}
@@ -877,24 +922,26 @@ class Output:
         obj = self.get_obj(filter_info=kwargs.get("filter_info"))
         if not obj:
             obj = self.odb.get(self.hash_info)
-        self._set_obj_names(obj)
 
-        return {None: {obj}}
+        return {None: set(self._named_obj_ids(obj))}
 
-    def _set_obj_names(self, obj):
-        obj.name = str(self)
+    def _named_obj_ids(self, obj):
+        name = str(self)
+        obj.hash_info.obj_name = name
+        yield obj.hash_info
         if isinstance(obj, Tree):
             for key, entry_obj in obj:
-                entry_obj.name = self.fs.sep.join([obj.name, *key])
+                entry_obj.hash_info.obj_name = self.fs.sep.join([name, *key])
+                yield entry_obj.hash_info
 
     def get_used_external(
         self, **kwargs
-    ) -> Dict[Optional["ObjectDB"], Set["HashFile"]]:
+    ) -> Dict[Optional["ObjectDB"], Set["HashInfo"]]:
         if not self.use_cache or not self.stage.is_repo_import:
             return {}
 
         (dep,) = self.stage.deps
-        return dep.get_used_objs()
+        return dep.get_used_objs(**kwargs)
 
     def _validate_output_path(self, path, stage=None):
         from dvc.dvcfile import is_valid_filename
@@ -957,6 +1004,18 @@ class Output:
     def fspath(self):
         return self.path_info.fspath
 
+    @property
+    def is_decorated(self) -> bool:
+        return self.is_metric or self.is_plot
+
+    @property
+    def is_metric(self) -> bool:
+        return bool(self.metric) or bool(self.live)
+
+    @property
+    def is_plot(self) -> bool:
+        return bool(self.plot)
+
 
 ARTIFACT_SCHEMA = {
     **CHECKSUMS_SCHEMA,
